import logging
import time
from os import path
from typing import Dict, Optional

from fastapi import APIRouter, Depends, Form, HTTPException, Request, status
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates
from pydantic_core._pydantic_core import ValidationError
from sqlalchemy.orm import Session
from typing_extensions import Annotated

from chanjo2 import __version__
from chanjo2.auth import get_current_user
from chanjo2.constants import DEFAULT_COVERAGE_LEVEL
from chanjo2.dbutil import get_session
from chanjo2.demo import DEMO_COVERAGE_QUERY_FORM
from chanjo2.meta.handle_report_contents import get_report_data
from chanjo2.models.pydantic_models import Builds, IntervalType, ReportQuery


def get_templates_path() -> str:
    """Returns the absolute path to the templates folder of this app."""
    APP_ROOT: str = path.abspath(path.join(path.dirname(__file__), ".."))
    return path.join(APP_ROOT, "templates")


LOG = logging.getLogger(__name__)
templates: Jinja2Templates = Jinja2Templates(directory=get_templates_path())
router = APIRouter()


@router.get("/report/demo", response_class=HTMLResponse)
async def demo_report(request: Request, db: Session = Depends(get_session)):
    """Return a demo coverage report over a list of genes for a list of samples."""

    report_query = ReportQuery.as_form(DEMO_COVERAGE_QUERY_FORM)
    report_content: Dict = get_report_data(query=report_query, session=db)
    return templates.TemplateResponse(
        request=request,
        name="report.html",
        context={
            "software_version": __version__,
            "levels": report_content["levels"],
            "extras": report_content["extras"],
            "sex_rows": report_content["sex_rows"],
            "completeness_rows": report_content["completeness_rows"],
            "default_level_completeness_rows": report_content[
                "default_level_completeness_rows"
            ],
            "interval_type": report_query.interval_type.value,
            "errors": report_content["errors"],
        },
    )


@router.post("/report", response_class=HTMLResponse)
async def report(
    request: Request,
    build=Annotated[Builds, Form(...)],
    samples=Annotated[str, Form(...)],
    interval_type=Annotated[IntervalType, Form(...)],
    completeness_thresholds=Annotated[Optional[str], Form(None)],
    ensembl_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_symbols=Annotated[Optional[str], Form(None)],
    case_display_name=Annotated[Optional[str], Form(None)],
    panel_name=Annotated[Optional[str], Form("Custom panel")],
    default_level=Annotated[Optional[int], Form(DEFAULT_COVERAGE_LEVEL)],
    db: Session = Depends(get_session),
    user: dict = Depends(get_current_user),
):
    """Return a coverage report over a list of genes for a list of samples."""
    start_time = time.time()
    try:
        report_query = ReportQuery.as_form(await request.form())
    except ValidationError as ve:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=ve.json(),
        )

    report_content: dict = get_report_data(query=report_query, session=db)
    LOG.debug(f"Time to compute stats: {time.time() - start_time} seconds.")
    return templates.TemplateResponse(
        request=request,
        name="report.html",
        context={
            "software_version": __version__,
            "levels": report_content["levels"],
            "extras": report_content["extras"],
            "sex_rows": report_content["sex_rows"],
            "completeness_rows": report_content["completeness_rows"],
            "default_level_completeness_rows": report_content[
                "default_level_completeness_rows"
            ],
            "interval_type": report_query.interval_type.value,
            "errors": report_content["errors"],
        },
    )



import logging
from typing import Dict, List, Optional, Union

from sqlalchemy import delete, func, or_, text
from sqlalchemy.orm import Session, query
from sqlalchemy.sql.expression import Delete

from chanjo2.models import SQLExon, SQLGene, SQLTranscript
from chanjo2.models.pydantic_models import Builds, TranscriptBase, TranscriptTag

LOG = logging.getLogger(__name__)


def delete_intervals_for_build(
    db: Session, interval_type: Union[SQLGene, SQLTranscript, SQLExon], build: Builds
) -> None:
    """Delete intervals from a given table by specifying a genome build."""
    statement: Delete = delete(interval_type).where(interval_type.build == build)
    db.execute(statement)


def count_intervals_for_build(
    db: Session, interval_type: Union[SQLGene, SQLTranscript, SQLExon], build: Builds
) -> int:
    """Count intervals in table by specifying a genome build."""
    return db.query(interval_type).where(interval_type.build == build).count()


def _filter_intervals_by_build(
    intervals: query.Query,
    interval_type: Union[SQLGene, SQLTranscript, SQLExon],
    build: Builds,
) -> query.Query:
    """Filter intervals by genome build."""
    return intervals.filter(interval_type.build == build)


def bulk_insert_genes(db: Session, genes: List[SQLGene]):
    """Bulk insert genes into the database."""
    db.bulk_save_objects(genes)
    db.commit()


def get_genes(
    db: Session,
    build: Builds,
    ensembl_ids: Optional[List[str]],
    hgnc_ids: Optional[List[int]],
    hgnc_symbols: Optional[List[str]],
    limit: Optional[int],
) -> List[SQLGene]:
    """Return genes according to specified fields."""
    genes: query.Query = db.query(SQLGene)
    if ensembl_ids:
        ensembl_ids_placeholder = ", ".join(f"'{e}'" for e in ensembl_ids)
        genes = genes.filter(
            text(
                f"""
                EXISTS (
                    SELECT 1
                    FROM json_each(genes.ensembl_ids)
                    WHERE value IN ({ensembl_ids_placeholder})
                )
                """
            )
        )
    elif hgnc_ids:
        genes: query.Query = genes.filter(SQLGene.hgnc_id.in_(hgnc_ids))
    elif hgnc_symbols:
        genes: query.Query = genes.filter(SQLGene.hgnc_symbol.in_(hgnc_symbols))

    genes: query.Query = _filter_intervals_by_build(
        intervals=genes, interval_type=SQLGene, build=build
    )
    if limit:
        return genes.limit(limit).all()
    return genes.all()


def bulk_insert_transcripts(db: Session, transcripts: List[TranscriptBase]):
    """Bulk insert transcripts into the database."""
    db.bulk_save_objects(transcripts)
    db.commit()


def get_hgnc_gene(db: Session, build: Builds, hgnc_id: int) -> SQLGene:
    """Return a gene object by its HGNC ID."""
    gene_query: query.Query = (
        db.query(SQLGene)
        .filter(SQLGene.hgnc_id == hgnc_id)
        .filter(SQLGene.build == build)
    )
    return gene_query.first()


def _filter_transcripts_by_tag(
    transcripts: query.Query, transcript_tags: List[TranscriptTag] = []
) -> query.Query:
    """Return transcripts which contain one or more RefSeq tag."""

    not_null_filters = []
    for tag in transcript_tags:
        not_null_filters.append(getattr(SQLTranscript, tag).isnot(None))
    return transcripts.filter(or_(*not_null_filters))


def set_sql_intervals(
    db: Session,
    interval_type: Union[SQLExon, SQLGene, SQLTranscript],
    genes: List[SQLGene],
    transcript_tags=Optional[List[TranscriptTag]],
) -> List[Union[SQLGene, SQLTranscript, SQLExon]]:
    """If SQL intervals are genes return them as they are, otherwise return genes, transcripts or exons."""
    if interval_type == SQLGene:
        sql_intervals: List[SQLGene] = genes
    else:
        sql_intervals: List[Union[SQLTranscript, SQLExon]] = get_gene_intervals(
            db=db,
            build=genes[0].build,
            interval_type=interval_type,
            ensembl_ids=None,
            hgnc_ids=None,
            hgnc_symbols=None,
            ensembl_gene_ids=[
                ensembl_id for gene in genes for ensembl_id in gene.ensembl_ids
            ],
            limit=None,
            transcript_tags=transcript_tags,
        )
    return sql_intervals


def get_gene_intervals(
    db: Session,
    build: Builds,
    interval_type: Union[SQLTranscript, SQLExon],
    ensembl_ids: Optional[List[str]] = [],
    hgnc_ids: Optional[List[int]] = [],
    hgnc_symbols: Optional[List[str]] = [],
    ensembl_gene_ids: Optional[List[str]] = [],
    limit: Optional[int] = None,
    transcript_tags: Optional[List[TranscriptTag]] = [],
) -> List[Union[SQLTranscript, SQLExon]]:
    """Retrieve transcripts or exons from a list of genes."""

    intervals = db.query(interval_type).filter(interval_type.build == build)

    def get_ensembl_gene_ids_from_gene_filter(
        filter_value: List[Union[str, int]], filter_column: str
    ) -> List[str]:
        """Helper function to get ensembl_gene_ids from either hgnc_ids or hgnc_symbols."""
        genes = (
            db.query(SQLGene.ensembl_ids).filter(filter_column.in_(filter_value)).all()
        )
        return [ensembl_id for gene in genes for ensembl_id in gene.ensembl_ids]

    if ensembl_ids:
        intervals: query.Query = intervals.filter(
            interval_type.ensembl_id.in_(ensembl_ids)
        )
    elif hgnc_ids:
        ensembl_gene_ids = get_ensembl_gene_ids_from_gene_filter(
            hgnc_ids, SQLGene.hgnc_id
        )
    elif hgnc_symbols:
        ensembl_gene_ids = get_ensembl_gene_ids_from_gene_filter(
            hgnc_symbols, SQLGene.hgnc_symbol
        )
    if ensembl_gene_ids:
        intervals = intervals.filter(
            interval_type.ensembl_gene_id.in_(ensembl_gene_ids)
        )

    if interval_type == SQLTranscript and transcript_tags:
        intervals = _filter_transcripts_by_tag(
            transcripts=intervals, transcript_tags=transcript_tags
        )

    if limit:
        return intervals.limit(limit).all()

    return intervals.all()


def bulk_insert_exons(db: Session, exons: List[SQLExon]) -> None:
    """Bulk insert exons into the database."""
    db.bulk_save_objects(exons)
    db.commit()


def get_interval_counts(db: Session) -> Dict:
    counts = {}
    for build in Builds.get_enum_values():
        counts[build] = {
            "number_of_genes": db.query(func.count(SQLGene.id))
            .filter(SQLGene.build == build)
            .scalar(),
            "number_of_transcripts": db.query(func.count(SQLTranscript.id))
            .filter(SQLTranscript.build == build)
            .scalar(),
            "number_of_exons": db.query(func.count(SQLExon.id))
            .filter(SQLExon.build == build)
            .scalar(),
        }
    return counts



from typing import Iterator, List, Tuple

CHROM_INDEX = 0
START_INDEX = 1
STOP_INDEX = 2


def resource_lines(file_path: str) -> Iterator[str]:
    resource = open(file_path, "r", encoding="utf-8")
    lines: List = resource.readlines()
    lines: List = [line.rstrip("\n") for line in lines]
    return iter(lines)


def bed_file_interval_id_coords(
    file_path: str,
) -> List[Tuple[str, Tuple[str, int, int]]]:
    """Parses a bed file and returns interval ID (gene or interval name) and coordinates."""

    with open(file_path, "r") as bed_file:
        bed_file_contents: List[List[str]] = [
            line.rstrip().replace("chr", "").split("\t")
            for line in bed_file
            if line.startswith("#") is False
        ]

    interval_id_coords = []

    for interval in bed_file_contents:
        interval_id: str = (
            "_".join(interval[STOP_INDEX + 1 : len(interval)])
            if len(interval) > STOP_INDEX
            else f"{CHROM_INDEX}:{START_INDEX}-{STOP_INDEX}"
        )
        interval_coords: Tuple[str, int, int] = (
            interval[CHROM_INDEX],
            int(interval[START_INDEX]),
            int(interval[STOP_INDEX]),
        )
        interval_id_coords.append((interval_id, interval_coords))

    return interval_id_coords


def sort_interval_ids_coords(
    interval_ids_coords: List[Tuple[str, Tuple[str, int, int]]],
) -> List[Tuple[str, Tuple[str, int, int]]]:
    """Sort intervals and their IDs by chrom, start and stop positions."""
    return sorted(
        interval_ids_coords,
        key=lambda interval_coord: (
            interval_coord[1][CHROM_INDEX],
            interval_coord[1][START_INDEX],
            interval_coord[1][STOP_INDEX],
        ),
    )



import logging
import tempfile
from statistics import mean
from typing import Dict, List, Optional, Tuple, Union

from chanjo2.meta.handle_bed import sort_interval_ids_coords
from chanjo2.meta.handle_completeness_stats import (
    get_completeness_stats,
    get_d4tools_intervals_completeness,
)
from chanjo2.meta.handle_coverage_stats import (
    get_chromosomes_prefix,
    get_d4tools_chromosome_mean_coverage,
    get_d4tools_intervals_coverage,
    get_d4tools_intervals_mean_coverage,
)
from chanjo2.models import SQLExon, SQLGene, SQLTranscript
from chanjo2.models.pydantic_models import ReportQuerySample, Sex

LOG = logging.getLogger(__name__)


def set_interval_ids_coords(
    sql_intervals: List[Union[SQLGene, SQLTranscript, SQLExon]],
) -> List[Tuple[str, Tuple[str, int, int]]]:
    """Returns tuples with an ensembl_id and coordinates from a list of SQL intervals."""

    if not sql_intervals:
        return []
    if isinstance(sql_intervals[0], SQLGene):
        return [
            (ensembl_id, (interval.chromosome, interval.start, interval.stop))
            for interval in sql_intervals
            for ensembl_id in interval.ensembl_ids
        ]
    else:
        return [
            (interval.ensembl_id, (interval.chromosome, interval.start, interval.stop))
            for interval in sql_intervals
        ]


def get_report_sample_interval_coverage(
    d4_file_path: str,
    sample_name: str,
    gene_ids_mapping: Dict[str, dict],
    sql_intervals: List[Union[SQLGene, SQLTranscript, SQLExon]],
    completeness_thresholds: List[Optional[int]],
    default_threshold: int,
    report_data: dict,
) -> None:
    """Compute stats to populate a coverage report for one sample."""

    # Compute intervals coverage completeness
    interval_ids_coords: List[Tuple[str, Tuple[str, int, int]]] = (
        set_interval_ids_coords(sql_intervals=sql_intervals)
    )
    interval_ids_coords = sort_interval_ids_coords(interval_ids_coords)

    chrom_prefix: str = get_chromosomes_prefix(d4_file_path)

    # Compute intervals coverage
    intervals_coverage: List[float] = get_d4tools_intervals_mean_coverage(
        d4_file_path=d4_file_path,
        interval_ids_coords=interval_ids_coords,
        chrom_prefix=chrom_prefix,
    )
    completeness_row_dict: dict = {"mean_coverage": mean(intervals_coverage)}

    # Compute intervals coverage completeenss
    intervals_coverage_completeness: Dict[str, dict] = get_completeness_stats(
        d4_file_path=d4_file_path,
        thresholds=completeness_thresholds,
        interval_ids_coords=interval_ids_coords,
        chrom_prefix=chrom_prefix,
    )

    interval_ids = set()
    thresholds_dict = {threshold: [] for threshold in completeness_thresholds}
    incomplete_coverages_rows: List[Tuple[int, str, str, str, float]] = []
    nr_intervals_covered_under_custom_threshold: int = 0
    genes_covered_under_custom_threshold = set()

    for interval in sql_intervals:

        if hasattr(interval, "ensembl_ids"):
            ensembl_ids = interval.ensembl_ids
        else:
            ensembl_ids = [interval.ensembl_id]

        for ensembl_id in ensembl_ids:

            if ensembl_id in interval_ids:
                continue
            for threshold in completeness_thresholds:
                interval_coverage_at_threshold: float = intervals_coverage_completeness[
                    ensembl_id
                ][threshold]
                thresholds_dict[threshold].append(interval_coverage_at_threshold)

                # Collect intervals which are not completely covered at the custom threshold
                if (
                    threshold == default_threshold
                    and interval_coverage_at_threshold < 1
                ):
                    nr_intervals_covered_under_custom_threshold += 1
                    interval_ensembl_gene: str = (
                        ensembl_id
                        if ensembl_id.startswith("ENSG")
                        else interval.ensembl_gene_id
                    )
                    interval_hgnc_id: int = gene_ids_mapping[interval_ensembl_gene][
                        "hgnc_id"
                    ]
                    interval_hgnc_symbol: str = gene_ids_mapping[interval_ensembl_gene][
                        "hgnc_symbol"
                    ]
                    genes_covered_under_custom_threshold.add(interval_hgnc_symbol)
                    incomplete_coverages_rows.append(
                        (
                            interval_hgnc_symbol,
                            interval_hgnc_id,
                            ensembl_id,
                            (
                                {
                                    "mane_select": interval.refseq_mane_select,
                                    "mane_plus_clinical": interval.refseq_mane_plus_clinical,
                                    "mrna": interval.refseq_mrna,
                                }
                                if isinstance(interval, SQLTranscript)
                                else {}
                            ),
                            sample_name,
                            round(interval_coverage_at_threshold * 100, 2),
                        )
                    )

            interval_ids.add(ensembl_id)

    for threshold in completeness_thresholds:
        if thresholds_dict[threshold]:
            completeness_row_dict[f"completeness_{threshold}"] = round(
                mean(thresholds_dict[threshold]) * 100, 2
            )

    report_data["completeness_rows"].append((sample_name, completeness_row_dict))
    report_data["incomplete_coverage_rows"] += incomplete_coverages_rows
    if interval_ids_coords:
        fully_covered_intervals_percent = round(
            100
            * (len(interval_ids) - nr_intervals_covered_under_custom_threshold)
            / len(interval_ids),
            2,
        )
        report_data["default_level_completeness_rows"].append(
            (
                sample_name,
                fully_covered_intervals_percent,
                f"{nr_intervals_covered_under_custom_threshold}/{len(interval_ids)}",
                genes_covered_under_custom_threshold,
            )
        )


def predict_sex(x_cov: float, y_cov: float) -> str:
    """Return predict sex based on sex chromosomes coverage - this code is taken from the old chanjo."""
    if y_cov == 0:
        return Sex.FEMALE.value
    else:
        ratio: float = x_cov / y_cov
        if x_cov == 0 or (ratio > 12 and ratio < 100):
            return Sex.UNKNOWN.value
        elif ratio <= 12:
            # this is the entire prediction, it's usually very obvious
            return Sex.MALE.value
        else:
            # the few reads mapping to the Y chromosomes are artifacts
            return Sex.FEMALE.value


def get_samples_sex_metrics(d4_file_path: str) -> Dict:
    """Compute coverage over sex chromosomes and predicted sex."""

    chrom_prefix: str = get_chromosomes_prefix(d4_file_path)

    sex_chroms_coverage: List[Tuple[str, float]] = get_d4tools_chromosome_mean_coverage(
        d4_file_path=d4_file_path, chromosomes=[f"{chrom_prefix}X", f"{chrom_prefix}Y"]
    )

    return {
        "x_coverage": round(sex_chroms_coverage[0][1], 1),
        "y_coverage": round(sex_chroms_coverage[1][1], 1),
        "predicted_sex": predict_sex(
            x_cov=sex_chroms_coverage[0][1], y_cov=sex_chroms_coverage[1][1]
        ),
    }


def get_gene_overview_stats(
    sql_intervals: List[SQLTranscript],
    samples: List[ReportQuerySample],
    completeness_thresholds: List[int],
) -> Dict[str, list]:
    """Returns stats to be included in the gene overview page."""
    interval_ids_coords: List[Tuple[str, Tuple[str, int, int]]] = (
        set_interval_ids_coords(sql_intervals=sql_intervals)
    )
    interval_ids_coords = tuple(
        sort_interval_ids_coords(set(interval_ids_coords))
    )  # removes duplicates and orders intervals by chromosome, start and stop
    transcripts_stats = {interval_id: [] for interval_id, _ in interval_ids_coords}

    chrom_prefix: str = get_chromosomes_prefix(
        samples[0].coverage_file_path
    )  # Assume they are all using the same reference

    # create a temp bed file containing transcripts coordinates
    bed_lines = [
        f"{chrom_prefix}{coords[0]}\t{coords[1]}\t{coords[2]}"
        for _, coords in interval_ids_coords
    ]
    temp_bed_file = tempfile.NamedTemporaryFile()
    with open(temp_bed_file.name, "w") as intervals_bed:
        intervals_bed.write("\n".join(bed_lines))
        intervals_bed.flush()

    for sample in samples:
        transcripts_coverage = get_d4tools_intervals_coverage(
            d4_file_path=sample.coverage_file_path, bed_file_path=temp_bed_file.name
        )
        transcripts_completeness = get_d4tools_intervals_completeness(
            d4_file_path=sample.coverage_file_path,
            bed_file_path=temp_bed_file.name,
            completeness_thresholds=completeness_thresholds,
        )
        for idx, transcripts_coords in enumerate(interval_ids_coords):
            append_tuple = (
                sample.name,
                transcripts_coverage[idx],
                transcripts_completeness[idx],
            )
            transcripts_stats[transcripts_coords[0]].append(append_tuple)

    return transcripts_stats



import logging
from collections import OrderedDict
from statistics import mean
from typing import Dict, List, Optional, Tuple, Union

from sqlalchemy.orm import Session

from chanjo2 import __version__
from chanjo2.crud.intervals import get_genes, get_hgnc_gene, set_sql_intervals
from chanjo2.meta.handle_d4 import (
    get_gene_overview_stats,
    get_report_sample_interval_coverage,
    get_samples_sex_metrics,
)
from chanjo2.models import SQLExon, SQLGene, SQLTranscript
from chanjo2.models.pydantic_models import (
    Builds,
    GeneReportForm,
    IntervalType,
    ReportQuery,
    ReportQuerySample,
    SampleSexRow,
    TranscriptTag,
)

LOG = logging.getLogger(__name__)
INTERVAL_TYPE_SQL_TYPE: Dict[IntervalType, Union[SQLGene, SQLTranscript, SQLExon]] = {
    IntervalType.GENES: SQLGene,
    IntervalType.TRANSCRIPTS: SQLTranscript,
    IntervalType.EXONS: SQLExon,
}

#### Functions used by all reports ####


def get_mean(float_list: List[float], round_by: int = 2) -> Union[float, str]:
    """Return the mean value from a list of floating point numbers, or a string when the value can't be converted to number."""
    if float_list:
        mean_value = round(mean(float_list), round_by)
    else:
        mean_value = "NA"

    return mean_value if str(mean_value).split(".")[0].isdigit() else str(mean_value)


def get_ordered_levels(threshold_levels: List[int]) -> OrderedDict:
    """Returns the coverage threshold levels as an ordered dictionary."""
    report_levels = OrderedDict()
    for threshold in sorted(threshold_levels):
        report_levels[threshold] = f"completeness_{threshold}"
    return report_levels


def set_samples_coverage_files(session: Session, samples: List[ReportQuerySample]):
    """Set path to coverage file for each sample in the samples list."""

    for sample in samples:
        if sample.coverage_file_path:  # if path to d4 is provided in the query
            continue
        else:  # fetch it from the database
            sql_sample: SQLSample = get_sample(db=session, sample_name=sample.name)
            sample.coverage_file_path = sql_sample.coverage_file_path


def _serialize_sample(sample: ReportQuerySample) -> Dict[str, str]:
    """Convert a ReportQuerySample object to an easily serialized dictionary."""
    return {
        "name": sample.name,
        "case_name": sample.case_name,
        "coverage_file_path": sample.coverage_file_path,
        "analysis_date": str(sample.analysis_date),
    }


def get_report_data(
    query: ReportQuery, session: Session, is_overview: Optional[bool] = False
) -> Dict:
    """Return the information that will be displayed in the coverage report or in the genes overview report.."""

    set_samples_coverage_files(session=session, samples=query.samples)

    if any([query.ensembl_gene_ids, query.hgnc_gene_ids, query.hgnc_gene_symbols]):
        genes: List[SQLGene] = get_genes(
            db=session,
            build=query.build,
            ensembl_ids=query.ensembl_gene_ids,
            hgnc_ids=query.hgnc_gene_ids,
            hgnc_symbols=query.hgnc_gene_symbols,
            limit=None,
        )
    else:
        genes = []

    data: Dict = {
        "levels": get_ordered_levels(threshold_levels=query.completeness_thresholds),
        "extras": {
            "panel_name": query.panel_name,
            "default_level": query.default_level,
            "interval_type": query.interval_type.value,
            "case_name": query.case_display_name,
            "hgnc_gene_ids": [gene.hgnc_id for gene in genes]
            or query.hgnc_gene_ids
            or query.hgnc_gene_symbols
            or query.ensembl_gene_ids,
            "build": query.build.value,
            "completeness_thresholds": query.completeness_thresholds,
            "samples": [_serialize_sample(sample) for sample in query.samples],
        },
        "completeness_rows": [],
        "incomplete_coverage_rows": [],
        "default_level_completeness_rows": [],
    }

    # Add coverage_report - specific data
    data["sex_rows"] = get_report_sex_rows(samples=query.samples)

    data["errors"] = [
        get_missing_genes_from_db(
            sql_genes=genes,
            ensembl_ids=query.ensembl_gene_ids,
            hgnc_ids=query.hgnc_gene_ids,
            hgnc_symbols=query.hgnc_gene_symbols,
        )
    ]

    gene_ids_mapping: Dict[str, dict] = {
        ensembl_id: {"hgnc_id": gene.hgnc_id, "hgnc_symbol": gene.hgnc_symbol}
        for gene in genes
        for ensembl_id in gene.ensembl_ids
    }

    sql_intervals: list = []
    if gene_ids_mapping:
        sql_intervals = set_sql_intervals(
            db=session,
            interval_type=INTERVAL_TYPE_SQL_TYPE[query.interval_type],
            genes=genes,
            transcript_tags=[
                TranscriptTag.REFSEQ_MANE_PLUS_CLINICAL,
                TranscriptTag.REFSEQ_MANE_SELECT,
                TranscriptTag.REFSEQ_MRNA,
            ],
        )

    for sample in query.samples:
        get_report_sample_interval_coverage(
            d4_file_path=sample.coverage_file_path,
            sample_name=sample.name,
            gene_ids_mapping=gene_ids_mapping,
            sql_intervals=sql_intervals,
            completeness_thresholds=(
                [query.default_level] if is_overview else query.completeness_thresholds
            ),
            default_threshold=query.default_level,
            report_data=data,
        )
    return data


#### Functions used to create coverage report data


def get_missing_genes_from_db(
    sql_genes: List[SQLGene],
    ensembl_ids: Optional[List[str]] = [],
    hgnc_ids: Optional[List[int]] = [],
    hgnc_symbols: Optional[List[str]] = [],
) -> Tuple[str, List[Union[str, int]]]:
    """Return queried genes that are not found in the database."""

    if ensembl_ids:
        sql_genes_ids = [gene.ensembl_id for gene in sql_genes]
    elif hgnc_ids:
        sql_genes_ids = [gene.hgnc_id for gene in sql_genes]
    else:
        sql_genes_ids = [gene.hgnc_symbol for gene in sql_genes]

    query_ids: List[Union[str, int]] = ensembl_ids or hgnc_ids or hgnc_symbols
    return "Gene IDs not found in the database", list(
        set(query_ids) - set(sql_genes_ids)
    )


def get_report_sex_rows(samples: List[ReportQuerySample]) -> List[Dict]:
    """Create and return the contents for the sample sex lines in the coverage report."""
    sample_sex_rows: D4FileList = []
    for sample in samples:
        sample_sex_metrics: Dict = get_samples_sex_metrics(
            d4_file_path=sample.coverage_file_path
        )

        sample_sex_row: SampleSexRow = SampleSexRow(
            **{
                "sample": sample.name,
                "case": sample.case_name,
                "analysis_date": sample.analysis_date,
                "predicted_sex": sample_sex_metrics["predicted_sex"],
                "x_coverage": sample_sex_metrics["x_coverage"],
                "y_coverage": sample_sex_metrics["y_coverage"],
            }
        )
        sample_sex_rows.append(sample_sex_row)
    return sample_sex_rows


#### Functions used to create a gene overview report ####


def get_gene_overview_coverage_stats(form_data: GeneReportForm, session: Session):
    """Returns coverage stats over the intervals of one gene for one or more samples."""

    gene_stats = {
        "software_version": __version__,
        "levels": get_ordered_levels(
            threshold_levels=form_data.completeness_thresholds
        ),
        "interval_type": form_data.interval_type.value,
        "transcript_coverage_stats": {},
    }

    set_samples_coverage_files(session=session, samples=form_data.samples)

    gene: SQLGene = get_hgnc_gene(
        build=form_data.build, hgnc_id=form_data.hgnc_gene_id, db=session
    )
    if gene is None:
        gene_stats["gene"] = {"hgnc_id": form_data.hgnc_gene_id}
        return gene_stats

    gene_stats["gene"] = gene
    transcripts_intervals = set_sql_intervals(
        db=session,
        interval_type=SQLTranscript,
        genes=[gene],
        transcript_tags=[],
    )
    exons_intervals = set_sql_intervals(db=session, interval_type=SQLExon, genes=[gene])
    sql_intervals = transcripts_intervals + exons_intervals

    samples_coverage_by_interval = get_gene_overview_stats(
        sql_intervals=sql_intervals,
        samples=form_data.samples,
        completeness_thresholds=form_data.completeness_thresholds,
    )

    for sql_interval in sql_intervals:
        interval_length: int = abs(sql_interval.stop - sql_interval.start)
        coords: str = (
            f"{sql_interval.chromosome}:{sql_interval.start}-{sql_interval.stop}"
        )

        if type(sql_interval) == SQLTranscript:
            gene_stats["transcript_coverage_stats"][sql_interval.ensembl_id] = {
                "interval_type": "transcript",
                "mane_select": sql_interval.refseq_mane_select,
                "mane_plus_clinical": sql_interval.refseq_mane_plus_clinical,
                "mrna": sql_interval.refseq_mrna,
                "stats": samples_coverage_by_interval[sql_interval.ensembl_id],
                "length": interval_length,
                "coordinates": coords,
                "exons": {},
            }
            continue

        gene_stats["transcript_coverage_stats"][sql_interval.ensembl_transcript_id][
            "exons"
        ][sql_interval.ensembl_id] = {
            "interval_type": "exon",
            "transcript_rank": int(sql_interval.rank_in_transcript),
            "stats": samples_coverage_by_interval[sql_interval.ensembl_id],
            "length": interval_length,
            "coordinates": coords,
        }

    return gene_stats


def get_mane_overview_coverage_stats(query: ReportQuery, session: Session) -> Dict:
    """Returns coverage stats over the MANE transcripts of a list of genes."""

    set_samples_coverage_files(session=session, samples=query.samples)
    genes = []
    if any([query.ensembl_gene_ids, query.hgnc_gene_ids, query.hgnc_gene_symbols]):
        genes: List[SQLGene] = get_genes(
            db=session,
            build=Builds.build_38,
            ensembl_ids=query.ensembl_gene_ids,
            hgnc_ids=query.hgnc_gene_ids,
            hgnc_symbols=query.hgnc_gene_symbols,
            limit=None,
        )

    gene_mappings = {}
    hgnc_gene_ids = []
    for gene in genes:
        hgnc_gene_ids.append(gene.hgnc_id)
        for ensembl_id in gene.ensembl_ids:
            gene_mappings[ensembl_id] = gene

    mane_stats = {
        "software_version": __version__,
        "levels": get_ordered_levels(threshold_levels=query.completeness_thresholds),
        "extras": {
            "hgnc_gene_ids": hgnc_gene_ids
            or query.hgnc_gene_ids
            or query.hgnc_gene_symbols
            or query.ensembl_gene_ids,
            "interval_type": query.interval_type.value,
            "default_level": query.default_level,
            "completeness_thresholds": query.completeness_thresholds,
            "samples": [_serialize_sample(sample) for sample in query.samples],
            "panel_name": query.panel_name,
        },
        "interval_type": IntervalType.TRANSCRIPTS,
        "mane_coverage_stats": [],
    }

    sql_intervals = []
    if genes:
        sql_intervals = set_sql_intervals(
            db=session,
            interval_type=SQLTranscript,
            genes=genes,
            transcript_tags=[
                TranscriptTag.REFSEQ_MANE_SELECT,
                TranscriptTag.REFSEQ_MANE_PLUS_CLINICAL,
            ],
        )

    mane_samples_coverage_stats_by_transcript = get_gene_overview_stats(
        sql_intervals=sql_intervals,
        samples=query.samples,
        completeness_thresholds=query.completeness_thresholds,
    )

    existing_transcripts = []

    for transcript in sql_intervals:
        transcript_dict = {
            "mane_select": transcript.refseq_mane_select,
            "mane_plus_clinical": transcript.refseq_mane_plus_clinical,
        }
        if transcript_dict in existing_transcripts:
            continue

        existing_transcripts.append(transcript_dict)
        gene_symbol: str = gene_mappings[transcript.ensembl_gene_id].hgnc_symbol
        data_dict: dict = {
            "gene": {
                "hgnc_id": gene_mappings[transcript.ensembl_gene_id].hgnc_id,
                "ensembl_ids": gene_mappings[transcript.ensembl_gene_id].ensembl_ids,
            },
            "transcript": transcript_dict,
            "stats": mane_samples_coverage_stats_by_transcript[transcript.ensembl_id],
        }
        mane_stats["mane_coverage_stats"].append((gene_symbol, data_dict))

    return mane_stats



import os

from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker
from sqlalchemy.pool import StaticPool

DEMO_DB = "sqlite://"
DEMO_CONNECT_ARGS = {"check_same_thread": False}

db_user = os.getenv("MYSQL_USER")
db_password = os.getenv("MYSQL_PASSWORD")
db_name = os.getenv("MYSQL_DATABASE_NAME")
host_name = os.getenv("MYSQL_HOST_NAME")
port_no = os.getenv("MYSQL_PORT")

if os.getenv("DEMO") or not db_name:
    mysql_url = DEMO_DB
    engine = create_engine(
        mysql_url,
        echo=True,
        connect_args=DEMO_CONNECT_ARGS,
        poolclass=StaticPool,
        future=True,
    )

else:
    if port_no == "":
        host = host_name
    else:
        host = ":".join([host_name, port_no])

    mysql_url = f"mysql://{db_user}:{db_password}@{host}/{db_name}"
    engine = create_engine(mysql_url, echo=True, future=True, pool_pre_ping=True)

SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine, future=True)
Base = declarative_base()


def get_session():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()



from typing import Callable, Dict, List

SAMPLE_NOT_FOUND: str = "One of more requested samples were not present in the database"
WRONG_COVERAGE_FILE_MSG: str = (
    "Coverage_file_path must be either an existing local file path or a URL"
)
WRONG_BED_FILE_MSG: str = "Provided intervals files is not a valid BED file"
MULTIPLE_PARAMS_NOT_SUPPORTED_MSG = "Interval query contains too many filter parameters. Please specify genome build and max one type of filter."
GENE_LISTS_NOT_SUPPORTED_MSG = (
    "Please provide either Ensembl gene IDs, HGNC gene IDS or HGNC gene symbols."
)
AMBIGUOUS_SAMPLES_INPUT = "Please provide either a name of a case or a list of samples."
HTTP_D4_COMPLETENESS_ERROR = "Completeness_thresholds must not be provided if any sample.coverage_file_path is a URL"

BUILD_37 = "GRCh37"
BUILD_38 = "GRCh38"
CHROMOSOME_NAME_37: str = "Chromosome Name"
CHROMOSOME_NAME_38: str = "Chromosome/scaffold name"
ENSEMBL_GENE_ID_37: str = "Ensembl Gene ID"
ENSEMBL_GENE_ID_38: str = "Gene stable ID"

GENES_FILE_HEADER: Dict[str, List[str]] = {
    BUILD_37: [
        CHROMOSOME_NAME_37,
        "Gene Start (bp)",
        "Gene End (bp)",
        ENSEMBL_GENE_ID_37,
        "HGNC symbol",
        "HGNC ID(s)",
    ],
    BUILD_38: [
        CHROMOSOME_NAME_38,
        "Gene start (bp)",
        "Gene end (bp)",
        ENSEMBL_GENE_ID_38,
        "HGNC symbol",
        "HGNC ID",
    ],
}

CHROMOSOMES = [str(i) for i in range(0, 23)] + ["X", "Y", "MT"]

TRANSCRIPTS_FILE_HEADER: Dict[str, List[str]] = {
    BUILD_37: [
        CHROMOSOME_NAME_37,
        ENSEMBL_GENE_ID_37,
        "Ensembl Transcript ID",
        "Transcript Start (bp)",
        "Transcript End (bp)",
        "RefSeq mRNA [e.g. NM_001195597]",
        "RefSeq mRNA predicted [e.g. XM_001125684]",
        "RefSeq ncRNA [e.g. NR_002834]",
    ],
    BUILD_38: [
        CHROMOSOME_NAME_38,
        ENSEMBL_GENE_ID_38,
        "Transcript stable ID",
        "Transcript start (bp)",
        "Transcript end (bp)",
        "RefSeq mRNA ID",
        "RefSeq mRNA predicted ID",
        "RefSeq ncRNA ID",
        "RefSeq match transcript (MANE Select)",
        "RefSeq match transcript (MANE Plus Clinical)",
    ],
}

EXONS_FILE_HEADER: Dict[str, List[str]] = {
    BUILD_37: [
        CHROMOSOME_NAME_37,
        ENSEMBL_GENE_ID_37,
        "Ensembl Transcript ID",
        "Ensembl Exon ID",
        "Exon Chr Start (bp)",
        "Exon Chr End (bp)",
        "5' UTR Start",
        "5' UTR End",
        "3' UTR Start",
        "3' UTR End",
        "Strand",
        "Exon Rank in Transcript",
    ],
    BUILD_38: [
        CHROMOSOME_NAME_38,
        ENSEMBL_GENE_ID_38,
        "Transcript stable ID",
        "Exon stable ID",
        "Exon region start (bp)",
        "Exon region end (bp)",
        "5' UTR start",
        "5' UTR end",
        "3' UTR start",
        "3' UTR end",
        "Strand",
        "Exon rank in transcript",
    ],
}

DEFAULT_COMPLETENESS_LEVELS = [10, 15, 20, 50, 100]
DEFAULT_COVERAGE_LEVEL = 10



import logging
from typing import Iterator, List, Optional, Union

from sqlalchemy.orm import Session
from tqdm import tqdm

from chanjo2.constants import (
    EXONS_FILE_HEADER,
    GENES_FILE_HEADER,
    TRANSCRIPTS_FILE_HEADER,
)
from chanjo2.crud.intervals import (
    bulk_insert_exons,
    bulk_insert_genes,
    bulk_insert_transcripts,
    count_intervals_for_build,
    delete_intervals_for_build,
)
from chanjo2.meta.handle_bed import resource_lines
from chanjo2.models import SQLExon, SQLGene, SQLTranscript
from chanjo2.models.pydantic_models import (
    Builds,
    ExonBase,
    IntervalType,
    TranscriptBase,
)

LOG = logging.getLogger(__name__)
MAX_NR_OF_RECORDS = 10_000
CHROM_SEPARATOR: str = "[success]"


def update_interval_table(
    interval_type: IntervalType,
    build: Builds,
    file_path: Optional[str],
    session: Session,
) -> None:
    """This function is runned in background and is responsible for updating a specific interval table of the database."""

    nlines: int = sum(1 for _ in open(file_path))
    interval_lines: Iterator[str] = resource_lines(file_path=file_path)

    if interval_type == IntervalType.GENES:
        update_genes(build=build, lines=interval_lines, nlines=nlines, session=session)
    elif interval_type == IntervalType.TRANSCRIPTS:
        update_transcripts(
            build=build, lines=interval_lines, nlines=nlines, session=session
        )
    elif interval_type == IntervalType.EXONS:
        update_exons(build=build, lines=interval_lines, nlines=nlines, session=session)


def _replace_empty_cols(line: str, nr_expected_columns: int) -> List[Union[str, None]]:
    """Split line into columns, replacing empty columns with None values."""
    cols = [
        None if cell == "" else cell.replace("HGNC:", "") for cell in line.split("\t")
    ]

    # Make sure that expected nr of cols are returned if last cols are blank
    cols += [None] * (nr_expected_columns - len(cols))
    return cols


def update_genes(build: Builds, session: Session, lines: Iterator, nlines: int) -> None:
    """Loads genes into the database, replacing existing ones."""

    LOG.warning(f"Updating genes. Genome build --> {build.value}")

    header = next(lines).split("\t")
    expected_header = GENES_FILE_HEADER[build]
    if header[: len(expected_header)] == expected_header is False:
        raise ValueError(
            f"Ensembl genes file has an unexpected format:{header}. Expected format: {GENES_FILE_HEADER[build]}"
        )

    LOG.warning(f"Deleting genes, transcripts, exons in build {build.value}")
    delete_intervals_for_build(db=session, interval_type=SQLGene, build=build)

    genes_bulk: List[SQLGene] = []

    with tqdm(total=nlines - 1, desc="Processing gene lines", unit="line") as pbar:
        for line in lines:
            line = line.strip()
            if line == CHROM_SEPARATOR:
                continue
            items: List = _replace_empty_cols(
                line=line, nr_expected_columns=len(header)
            )
            if items == header:
                continue
            sql_gene = SQLGene(
                build=build,
                chromosome=items[0],
                start=int(items[1]),
                stop=int(items[2]),
                ensembl_ids=[items[3]],
                hgnc_symbol=items[4],
                hgnc_id=items[5],
            )

            genes_bulk.append(sql_gene)

            # Bulk insert when threshold is reached
            if len(genes_bulk) >= MAX_NR_OF_RECORDS:
                bulk_insert_genes(db=session, genes=genes_bulk)
                genes_bulk = []

            pbar.update(1)

    # Insert remaining genes
    if genes_bulk:
        bulk_insert_genes(db=session, genes=genes_bulk)

    pbar.close()

    nr_loaded_genes: int = count_intervals_for_build(
        db=session, interval_type=SQLGene, build=build
    )
    LOG.warning(f"{nr_loaded_genes} genes loaded into the database.")


def update_transcripts(
    build: Builds, session: Session, lines: [Iterator], nlines: int
) -> None:
    """Loads transcripts into the database."""

    LOG.warning(f"Updating transcripts. Genome build --> {build.value}")

    header = next(lines).split("\t")
    expected_header = TRANSCRIPTS_FILE_HEADER[build]
    if header[: len(expected_header)] == expected_header is False:
        raise ValueError(
            f"Ensembl transcripts file has an unexpected format:{header}. Expected format: {TRANSCRIPTS_FILE_HEADER[build]}"
        )

    LOG.warning(f"Deleting transcripts in build {build.value}")
    delete_intervals_for_build(db=session, interval_type=SQLTranscript, build=build)

    transcripts_bulk: List[TranscriptBase] = []

    with tqdm(
        total=nlines - 1, desc="Processing transcripts lines", unit="line"
    ) as pbar:
        for line in lines:
            line = line.strip()
            if line == CHROM_SEPARATOR:
                continue
            items: List = _replace_empty_cols(
                line=line, nr_expected_columns=len(header)
            )
            if items == header:
                continue
            transcript = SQLTranscript(
                chromosome=items[0],
                ensembl_gene_id=items[1],
                ensembl_id=items[2],
                start=int(items[3]),
                stop=int(items[4]),
                refseq_mrna=items[5],
                refseq_mrna_pred=items[6],
                refseq_ncrna=items[7],
                refseq_mane_select=items[8] if build == Builds.build_38 else None,
                refseq_mane_plus_clinical=(
                    items[9] if build == Builds.build_38 else None
                ),
                build=build,
            )

            transcripts_bulk.append(transcript)

            # Bulk insert when threshold is reached
            if len(transcripts_bulk) > MAX_NR_OF_RECORDS:
                bulk_insert_transcripts(db=session, transcripts=transcripts_bulk)
                transcripts_bulk = []

            pbar.update(1)

    # Insert remaining genes
    if transcripts_bulk:
        bulk_insert_transcripts(db=session, transcripts=transcripts_bulk)

    pbar.close()
    nr_loaded_transcripts: int = count_intervals_for_build(
        db=session, interval_type=SQLTranscript, build=build
    )
    LOG.warning(f"{nr_loaded_transcripts} transcripts loaded into the database.")


def update_exons(
    build: Builds, session: Session, lines: [Iterator], nlines: int
) -> None:
    """Loads exons into the database."""

    LOG.warning(f"Updating exons. Genome build --> {build.value}")

    header = next(lines).split("\t")
    expected_header = EXONS_FILE_HEADER[build]
    if header[: len(expected_header)] == expected_header is False:
        raise ValueError(
            f"Ensembl exons file has an unexpected format:{header}. Expected format: {EXONS_FILE_HEADER[build]}"
        )

    LOG.warning(f"Deleting exons in build {build.value}")
    delete_intervals_for_build(db=session, interval_type=SQLExon, build=build)

    exons_bulk: List[ExonBase] = []

    with tqdm(total=nlines - 1, desc="Processing exons lines", unit="line") as pbar:
        for line in lines:
            line = line.strip()
            if line == CHROM_SEPARATOR:
                continue
            items: List = _replace_empty_cols(
                line=line, nr_expected_columns=len(header)
            )
            if items == header:
                continue
            exon = SQLExon(
                chromosome=items[0],
                ensembl_gene_id=items[1],
                ensembl_transcript_id=items[2],
                ensembl_id=items[3],
                start=int(items[4]),
                stop=int(items[5]),
                rank_in_transcript=int(items[-1]),
                build=build,
            )

            exons_bulk.append(exon)

            # Bulk insert when threshold is reached
            if len(exons_bulk) > MAX_NR_OF_RECORDS:
                bulk_insert_exons(db=session, exons=exons_bulk)
                exons_bulk = []

            pbar.update(1)

    # Insert remaining genes
    if exons_bulk:
        bulk_insert_exons(db=session, exons=exons_bulk)

    pbar.close()
    nr_loaded_exons: int = count_intervals_for_build(
        db=session, interval_type=SQLExon, build=build
    )
    LOG.warning(f"{nr_loaded_exons} exons loaded into the database.")



import json
import os
from datetime import datetime
from enum import Enum
from os.path import isfile
from typing import Dict, List, Optional, Union

import validators
from fastapi import HTTPException, status
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator
from starlette.datastructures import FormData

from chanjo2.constants import (
    DEFAULT_COMPLETENESS_LEVELS,
    DEFAULT_COVERAGE_LEVEL,
    GENE_LISTS_NOT_SUPPORTED_MSG,
    HTTP_D4_COMPLETENESS_ERROR,
    WRONG_COVERAGE_FILE_MSG,
)


def validate_url_and_completeness(
    d4_file: str, completeness_thresholds: Optional[List[int]]
):
    """Raise error if d4 file is HTTP file and completeness thresholds are present."""
    if not completeness_thresholds:
        return
    if is_valid_url(d4_file) is False:
        return
    raise HTTPException(
        status.HTTP_422_UNPROCESSABLE_ENTITY,
        detail=HTTP_D4_COMPLETENESS_ERROR,
    )


def is_valid_url(value: str) -> bool:
    """Makes sure that a string is formatted as an URL."""
    try:
        return bool(validators.url(value))
    except Exception:
        return False


def default_report_coverage_levels() -> List[int]:
    """Sets the coverage thresholds to be used for report metrics whenever a request doesn't contain 'completeness_thresholds' values."""
    if os.getenv("REPORT_COVERAGE_LEVELS"):
        return json.loads(os.getenv("REPORT_COVERAGE_LEVELS"))
    return DEFAULT_COMPLETENESS_LEVELS


class Builds(str, Enum):
    build_37 = "GRCh37"
    build_38 = "GRCh38"

    @staticmethod
    def get_enum_values() -> List[str]:
        """Returns the values of the available genome builds."""
        return [member.value for member in Builds]


class IntervalType(str, Enum):
    GENES = "genes"
    TRANSCRIPTS = "transcripts"
    EXONS = "exons"
    CUSTOM = "custom_intervals"

    @staticmethod
    def get_enum_values() -> List[str]:
        """Returns the values of the available interval types."""
        return [member.value for member in IntervalType]


class TranscriptTag(str, Enum):
    REFSEQ_MRNA = "refseq_mrna"
    REFSEQ_NCRNA = "refseq_ncrna"
    REFSEQ_MANE_SELECT = "refseq_mane_select"
    REFSEQ_MANE_PLUS_CLINICAL = "refseq_mane_plus_clinical"


class Sex(str, Enum):
    FEMALE = "female"
    MALE = "male"
    UNKNOWN = "unknown"


class IntervalBase(BaseModel):
    chromosome: str
    start: int
    stop: int


class Interval(IntervalBase):
    id: int


class GeneBase(IntervalBase):
    build: Builds
    ensembl_ids: List[str]
    hgnc_id: Optional[int]
    hgnc_symbol: Optional[str]

    model_config = ConfigDict(from_attributes=True)


class GeneQuery(BaseModel):
    build: Builds
    ensembl_ids: Optional[List[str]] = None
    hgnc_ids: Optional[List[int]] = None
    hgnc_symbols: Optional[List[str]] = None
    limit: Optional[int] = 100


class GeneIntervalQuery(GeneQuery):
    ensembl_gene_ids: Optional[List[str]] = None


class Gene(IntervalBase):
    id: int


class TranscriptBase(IntervalBase):
    build: Builds
    ensembl_id: str
    ensembl_gene_id: str
    refseq_mrna: Optional[str] = None
    refseq_mrna_pred: Optional[str] = None
    refseq_ncrna: Optional[str] = None
    refseq_mane_select: Optional[str] = None
    refseq_mane_plus_clinical: Optional[str] = None


class Transcript(TranscriptBase):
    id: int


class ExonBase(IntervalBase):
    build: Builds
    ensembl_id: str
    ensembl_gene_id: str
    ensembl_transcript_id: str
    rank_in_transcript: int


class Exon(IntervalBase):
    id: int


class IntervalCoverage(BaseModel):
    mean_coverage: float
    completeness: Optional[Dict] = Field(default_factory=dict)
    interval_id: Optional[str] = None
    interval_type: Optional[IntervalType] = IntervalType.CUSTOM


class GeneCoverage(IntervalCoverage):
    inner_intervals: List[IntervalCoverage] = []  # Transcripts or exons
    hgnc_id: Optional[int] = None
    hgnc_symbol: Optional[str] = None
    ensembl_gene_id: Optional[str] = None


class FileCoverageBaseQuery(BaseModel):
    coverage_file_path: str


class FileCoverageQuery(FileCoverageBaseQuery):
    chromosome: str
    start: Optional[int] = None
    end: Optional[int] = None
    completeness_thresholds: Optional[List[int]] = []
    coverage_file_path: str

    @field_validator("coverage_file_path", mode="after")
    def coverage_file_path_validator(cls, coverage_file_path):
        if isfile(coverage_file_path) or is_valid_url(coverage_file_path):
            return coverage_file_path
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail=WRONG_COVERAGE_FILE_MSG)

    @model_validator(mode="after")
    def coverage_path_completeness_validator(self):
        """Completeness computation is not supported for d4 files over HTTP."""
        validate_url_and_completeness(
            d4_file=self.coverage_file_path,
            completeness_thresholds=self.completeness_thresholds,
        )
        return self


class FileCoverageIntervalsFileQuery(FileCoverageBaseQuery):
    intervals_bed_path: str
    completeness_thresholds: Optional[List[int]] = []
    coverage_file_path: str

    @field_validator("coverage_file_path", mode="after")
    def coverage_file_path_validator(cls, coverage_file_path):
        if isfile(coverage_file_path) or is_valid_url(coverage_file_path):
            return coverage_file_path
        raise HTTPException(status.HTTP_404_NOT_FOUND, detail=WRONG_COVERAGE_FILE_MSG)

    @model_validator(mode="after")
    def coverage_path_completeness_validator(self):
        """Completeness computation is not supported for d4 files over HTTP."""
        validate_url_and_completeness(
            d4_file=self.coverage_file_path,
            completeness_thresholds=self.completeness_thresholds,
        )
        return self


## Coverage and  overview report - related models ###
class CoverageSummaryQuerySample(BaseModel):
    name: str
    coverage_file_path: str


class CoverageSummaryQuery(BaseModel):
    build: Builds
    samples: List[CoverageSummaryQuerySample]
    hgnc_gene_ids: List[int]
    coverage_threshold: int
    interval_type: IntervalType

    @model_validator(mode="after")
    def check_no_http_cov_files(self):
        """Completeness computation, which is performed downstream, is not supported for d4 files over HTTP."""
        for sample in self.samples:
            if is_valid_url(sample.coverage_file_path):
                raise HTTPException(
                    status.HTTP_422_UNPROCESSABLE_ENTITY,
                    detail=HTTP_D4_COMPLETENESS_ERROR,
                )
        return self


class ReportQuerySample(BaseModel):
    name: str
    coverage_file_path: Optional[str] = None
    case_name: Optional[str] = None
    analysis_date: Optional[datetime] = datetime.now()

    @field_validator("coverage_file_path", mode="after")
    def coverage_file_path_validator(cls, coverage_file_path):
        if isfile(coverage_file_path) or is_valid_url(coverage_file_path):
            return coverage_file_path
        raise ValueError(WRONG_COVERAGE_FILE_MSG)


class ReportQuery(BaseModel):
    build: Builds
    completeness_thresholds: Optional[List[int]] = default_report_coverage_levels()
    ensembl_gene_ids: Optional[List[str]] = None
    hgnc_gene_ids: Optional[List[int]] = None
    hgnc_gene_symbols: Optional[List[str]] = None
    interval_type: IntervalType
    default_level: int = DEFAULT_COVERAGE_LEVEL
    panel_name: Optional[str] = None
    case_display_name: Optional[str] = None
    samples: List[ReportQuerySample]

    @model_validator(mode="after")
    def coverage_paths_completeness_validator(self):
        """Completeness computation is not supported for d4 files over HTTP. Check each sample."""
        for sample in self.samples:
            validate_url_and_completeness(
                d4_file=sample.coverage_file_path,
                completeness_thresholds=self.completeness_thresholds,
            )
        return self

    @staticmethod
    def comma_sep_values_to_list(
        comma_sep_values: Optional[str], items_format: Union[str, int]
    ) -> Optional[List[Union[str, int]]]:
        """Helper function that formats list of strings or integers passed by a form as comma separated values."""
        if comma_sep_values is None:
            return
        if items_format == str:
            return [item.strip() for item in comma_sep_values.split(",")]
        else:
            return [int(item.strip()) for item in comma_sep_values.split(",")]

    @staticmethod
    def set_query_genes(form_data: FormData):
        """Helper function that collects form data from report page requests and sets the right gene IDs/values in the query."""
        query_genes = {
            "ensembl_gene_ids": [],
            "hgnc_gene_ids": [],
            "hgnc_gene_symbols": [],
        }

        for gene_ids_key in query_genes.keys():
            if bool(form_data.get(gene_ids_key)) is False:
                continue
            id_values: List[Union[str, int]] = [
                item.strip() for item in form_data.get(gene_ids_key).split(",")
            ]
            for value in id_values:
                if value.isdigit():
                    query_genes["hgnc_gene_ids"].append(value)
                elif value.startswith("ENSG"):
                    query_genes["ensembl_gene_ids"].append(value)
                else:
                    query_genes["hgnc_gene_symbols"].append(value)
        return query_genes

    @classmethod
    def as_form(cls, form_data: FormData) -> "ReportQuery":
        query_genes: dict = cls.set_query_genes(form_data)
        report_query: dict = {
            "build": form_data.get("build"),
            "completeness_thresholds": (
                cls.comma_sep_values_to_list(
                    comma_sep_values=form_data.get("completeness_thresholds"),
                    items_format=int,
                )
                if form_data.get("completeness_thresholds")
                else default_report_coverage_levels()
            ),
            "ensembl_gene_ids": query_genes["ensembl_gene_ids"],
            "hgnc_gene_ids": query_genes["hgnc_gene_ids"],
            "hgnc_gene_symbols": query_genes["hgnc_gene_symbols"],
            "interval_type": form_data.get("interval_type"),
            "default_level": (
                form_data.get("default_level")
                if form_data.get("default_level")
                else DEFAULT_COVERAGE_LEVEL
            ),
            "panel_name": form_data.get("panel_name"),
            "case_display_name": form_data.get("case_display_name"),
            "samples": form_data.get("samples"),
        }
        return cls(**report_query)

    @field_validator("samples", mode="before")
    def samples_validator(cls, sample_list):
        if isinstance(sample_list, str):
            return json.loads(sample_list.replace("'", '"'))
        return sample_list

    @model_validator(mode="before")
    def check_genes_lists(cls, values: dict):
        nr_provided_gene_lists: int = 0
        for gene_list in [
            values.get("ensembl_gene_ids"),
            values.get("hgnc_gene_ids"),
            values.get("hgnc_gene_symbols"),
        ]:
            if gene_list:
                nr_provided_gene_lists += 1
        if nr_provided_gene_lists > 1:
            raise ValueError(GENE_LISTS_NOT_SUPPORTED_MSG)
        return values


class GeneReportForm(BaseModel):
    build: Builds
    completeness_thresholds: Optional[List[int]] = default_report_coverage_levels()
    hgnc_gene_id: int
    default_level: int = DEFAULT_COVERAGE_LEVEL
    samples: List[ReportQuerySample]
    interval_type: IntervalType

    @model_validator(mode="after")
    def check_thresholds_not_with_url_d4(self):
        """Completeness computation is not supported for d4 files over HTTP."""
        if self.completeness_thresholds:
            for sample in self.samples:
                if is_valid_url(sample.coverage_file_path):
                    raise HTTPException(
                        status.HTTP_422_UNPROCESSABLE_ENTITY,
                        detail=HTTP_D4_COMPLETENESS_ERROR,
                    )
        return self

    @field_validator("samples", mode="before")
    def samples_validator(cls, sample_list):
        if isinstance(sample_list, str):
            return json.loads(sample_list.replace("'", '"'))
        return sample_list

    @field_validator("completeness_thresholds", mode="before")
    def coverage_thresholds_validator(cls, completeness_thresholds: str):
        thresholds: List[str] = completeness_thresholds.split(",")
        return [int(threshold.strip()) for threshold in thresholds]

    @field_validator("interval_type", mode="before")
    def interval_type_validator(cls, interval_type: IntervalType) -> IntervalType:
        """Make sure that gene stats include transcript data even if it's a WGS analysis."""
        if interval_type == IntervalType.GENES:
            return IntervalType.TRANSCRIPTS
        return interval_type


class SampleSexRow(BaseModel):
    sample: str
    case: Optional[str]
    analysis_date: datetime
    predicted_sex: str
    x_coverage: float
    y_coverage: float



import subprocess
import tempfile
from typing import List, Tuple

from chanjo2.constants import CHROMOSOMES

CHROM_INDEX = 0
START_INDEX = 1
STOP_INDEX = 2
STATS_MEAN_COVERAGE_INDEX = 3


def get_chromosomes_prefix(d4_file_path: str) -> str:
    """Extracts the prefix to be prepended to genomic intervals when calculating stats."""

    # SonarCloud: d4_file_path is validated upstream
    result = subprocess.run(
        ["d4tools", "view", "-g", d4_file_path],
        capture_output=True,
        text=True,
        check=True,
    )
    first_line = result.stdout.splitlines()[0] if result.stdout else ""

    if "chr" in first_line:
        return "chr"
    return ""


def get_d4tools_intervals_mean_coverage(
    d4_file_path: str, interval_ids_coords: List[Tuple[str, tuple]], chrom_prefix: str
) -> List[float]:
    """Return the mean value over a list of intervals of a d4 file."""

    if interval_ids_coords:
        bed_lines = [
            f"{chrom_prefix}{coords[CHROM_INDEX]}\t{coords[START_INDEX]}\t{coords[STOP_INDEX]}"
            for _, coords in interval_ids_coords
        ]
        # Write genomic intervals to a temporary file
        with tempfile.NamedTemporaryFile(mode="w") as intervals_bed:
            intervals_bed.write("\n".join(bed_lines))
            intervals_bed.flush()

            return get_d4tools_intervals_coverage(
                d4_file_path=d4_file_path, bed_file_path=intervals_bed.name
            )
    chromosomes_mean_cov = get_d4tools_chromosome_mean_coverage(
        d4_file_path=d4_file_path, chromosomes=CHROMOSOMES
    )

    return [chrom_cov[1] for chrom_cov in chromosomes_mean_cov]


def get_d4tools_intervals_coverage(
    d4_file_path: str, bed_file_path: str
) -> List[float]:
    """Return the coverage for intervals of a d4 file that are found in a bed file."""

    d4tools_stats_mean_cmd: str = subprocess.check_output(
        ["d4tools", "stat", "--region", bed_file_path, d4_file_path, "--stat", "mean"],
        text=True,
    )
    return [
        float(line.rstrip().split("\t")[3])
        for line in d4tools_stats_mean_cmd.splitlines()
    ]


def get_d4tools_chromosome_mean_coverage(
    d4_file_path: str, chromosomes=List[str]
) -> List[Tuple[str, float]]:
    """Return mean coverage over entire chromosomes."""

    chromosomes_stats_mean_cmd: List[str] = subprocess.check_output(
        ["d4tools", "stat", "-s" "mean", d4_file_path],
        text=True,
    ).splitlines()
    chromosomes_coverage: List[Tuple[str, float]] = []
    for line in chromosomes_stats_mean_cmd:
        stats_data: List[str] = line.split("\t")
        if stats_data[CHROM_INDEX] in chromosomes:
            chromosomes_coverage.append(
                (stats_data[CHROM_INDEX], float(stats_data[STATS_MEAN_COVERAGE_INDEX]))
            )
    return chromosomes_coverage



import logging
import os
from contextlib import asynccontextmanager
from typing import List, Tuple

from fastapi import FastAPI, status
from fastapi.staticfiles import StaticFiles

from chanjo2 import __version__
from chanjo2.dbutil import engine
from chanjo2.endpoints import coverage, intervals, overview, report
from chanjo2.logger import configure_log
from chanjo2.models.sql_models import Base
from chanjo2.populate_demo import load_demo_data

LOG = logging.getLogger(__name__)
APP_ROUTER_TAGS: List[Tuple] = [
    (intervals.router, "intervals"),
    (coverage.router, "coverage"),
    (report.router, "report"),
    (overview.router, "overview"),
]


def create_db_and_tables():
    Base.metadata.create_all(engine)


@asynccontextmanager
async def lifespan(app_: FastAPI):
    configure_log(app=app, log=LOG)
    LOG.info("Starting up...")
    await startup_db()
    yield
    LOG.info("Shutting down...")


app = FastAPI(lifespan=lifespan)


def configure_static(app):
    """Configure static folder."""
    exec_dir: str = os.path.dirname(__file__)
    static_abs_file_path: str = os.path.join(exec_dir, "static/")
    app.mount("/static", StaticFiles(directory=static_abs_file_path), name="static")


# include app router tags
for router, tag in APP_ROUTER_TAGS:
    app.include_router(
        router,
        tags=[tag],
        responses={status.HTTP_404_NOT_FOUND: {"description": "Not found"}},
    )


async def startup_db():
    # Create database tables
    create_db_and_tables()

    configure_static(app)

    if os.getenv("DEMO") or not os.getenv("MYSQL_DATABASE_NAME"):
        LOG.warning("Running a demo instance of Chanjo2")
        if await load_demo_data():
            LOG.info("Demo data loaded into database")


@app.get("/")
def heartbeat():
    return {"message": f"Chanjo2 v{__version__} is up and running!"}



from typing import Dict

from importlib_resources import files

from chanjo2.constants import BUILD_37, DEFAULT_COMPLETENESS_LEVELS

BASE_PATH: str = "chanjo2.demo"

GENE_PANEL_FILE: str = "109_green.bed"
D4_DEMO_FILE: str = "panelapp_109_example.d4"

# Paths
gene_panel_path = str(files(BASE_PATH).joinpath(GENE_PANEL_FILE))
d4_demo_path = str(files(BASE_PATH).joinpath(D4_DEMO_FILE))

DEMO_CASE: Dict[str, str] = {"name": "internal_id", "display_name": "643594"}

DEMO_SAMPLE: Dict[str, str] = {
    "name": "ADM1059A2",
    "display_name": "NA12882",
    "track_name": "ADM1059A2",
    "case_name": DEMO_CASE["name"],
    "coverage_file_path": d4_demo_path,
}

HTTP_SERVER_D4_file = "https://d4-format-testing.s3.us-west-1.amazonaws.com/hg002.d4"
DEMO_HGNC_GENE_SYMBOLS = ["MTHFR", "DHFR", "FOLR1", "SLC46A1", "LAMA1", "PIPPI6"]
DEMO_HGNC_IDS = [7436, 2861, 3791, 30521]
ANALYSIS_DATE = "2023-04-23T10:20:30.400+02:30"

DEMO_GENE_OVERVIEW_QUERY_FORM = {
    "build": BUILD_37,
    "completeness_thresholds": ",".join(
        [str(threshold) for threshold in DEFAULT_COMPLETENESS_LEVELS]
    ),
    "hgnc_gene_id": DEMO_HGNC_IDS[0],
    "default_level": "20",
    "samples": f"[{{'name': '{DEMO_SAMPLE['name']}', 'coverage_file_path': '{d4_demo_path}', 'case_name': '{DEMO_CASE['name']}', 'analysis_date': '{ANALYSIS_DATE}'}}]",
    "interval_type": "transcripts",
}

# HTTP FORM-like data for generating a demo coverage report
DEMO_COVERAGE_QUERY_FORM = {
    "build": BUILD_37,
    "samples": f"[{{'name': '{DEMO_SAMPLE['name']}', 'coverage_file_path': '{d4_demo_path}', 'case_name': '{DEMO_CASE['name']}', 'analysis_date': '{ANALYSIS_DATE}'}}]",
    "case_display_name": DEMO_CASE["name"],
    "gene_panel": "A test Panel 1.0",
    "interval_type": "transcripts",
    "ensembl_gene_ids": [],
    "hgnc_gene_ids": [],
    "hgnc_gene_symbols": ",".join(DEMO_HGNC_GENE_SYMBOLS),
    "default_level": "20",
    "completeness_thresholds": ",".join(
        [str(threshold) for threshold in DEFAULT_COMPLETENESS_LEVELS]
    ),
}

intervals = "intervals"
EXONS_37_FILE = "exons_37.tsv"
EXONS_37_FILE_PATH = str(files(BASE_PATH).joinpath(intervals).joinpath(EXONS_37_FILE))
EXONS_38_FILE = "exons_38.tsv"
EXONS_38_FILE_PATH = str(files(BASE_PATH).joinpath(intervals).joinpath(EXONS_38_FILE))
GENES_37_FILE = "genes_37.tsv"
GENES_37_FILE_PATH = str(files(BASE_PATH).joinpath(intervals).joinpath(GENES_37_FILE))
GENES_38_FILE = "genes_38.tsv"
GENES_38_FILE_PATH = str(files(BASE_PATH).joinpath(intervals).joinpath(GENES_38_FILE))
TRANSCRIPTS_37_FILE = "transcripts_37.tsv"
TRANSCRIPTS_37_FILE_PATH = str(
    files(BASE_PATH).joinpath(intervals).joinpath(TRANSCRIPTS_37_FILE)
)
TRANSCRIPTS_38_FILE = "transcripts_38.tsv"
TRANSCRIPTS_38_FILE_PATH = str(
    files(BASE_PATH).joinpath(intervals).joinpath(TRANSCRIPTS_38_FILE)
)



from typing import Iterator, List, Tuple

from sqlalchemy.orm import sessionmaker

from chanjo2.dbutil import get_session
from chanjo2.demo import (
    EXONS_37_FILE_PATH,
    EXONS_38_FILE_PATH,
    GENES_37_FILE_PATH,
    GENES_38_FILE_PATH,
    TRANSCRIPTS_37_FILE_PATH,
    TRANSCRIPTS_38_FILE_PATH,
)
from chanjo2.meta.handle_bed import resource_lines
from chanjo2.meta.handle_load_intervals import (
    update_exons,
    update_genes,
    update_transcripts,
)
from chanjo2.models.pydantic_models import Builds

db: sessionmaker = next(get_session())

BUILD_GENES_RESOURCE: List[Tuple[Builds, str]] = [
    (Builds.build_37, GENES_37_FILE_PATH),
    (Builds.build_38, GENES_38_FILE_PATH),
]
BUILD_TRANSCRIPTS_RESOURCE: List[Tuple[Builds, str]] = [
    (Builds.build_37, TRANSCRIPTS_37_FILE_PATH),
    (Builds.build_38, TRANSCRIPTS_38_FILE_PATH),
]
BUILD_EXONS_RESOURCE: List[Tuple[Builds, str]] = [
    (Builds.build_37, EXONS_37_FILE_PATH),
    (Builds.build_38, EXONS_38_FILE_PATH),
]


async def load_demo_data() -> None:
    """Loads demo data into the database of a demo instance of Chanjo2."""
    await load_demo_genes()
    await load_demo_transcripts()
    await load_demo_exons()


async def load_demo_genes() -> None:
    """Load test genes into the database."""

    for build, path in BUILD_GENES_RESOURCE:
        gene_lines: Iterator = resource_lines(path)
        nlines: int = sum(1 for _ in open(path))
        update_genes(build=build, session=db, lines=gene_lines, nlines=nlines)


async def load_demo_transcripts() -> None:
    """Load test transcripts into the database."""

    for build, path in BUILD_TRANSCRIPTS_RESOURCE:
        transcript_lines: Iterator = resource_lines(path)
        nlines: int = sum(1 for _ in open(path))
        update_transcripts(
            build=build, session=db, lines=transcript_lines, nlines=nlines
        )


async def load_demo_exons() -> None:
    """Load test exons into the database."""

    for build, path in BUILD_EXONS_RESOURCE:
        exon_lines: Iterator = resource_lines(path)
        nlines: int = sum(1 for _ in open(path))
        update_exons(build=build, session=db, lines=exon_lines, nlines=nlines)



import subprocess
import tempfile
from typing import Dict, List, Tuple

CHROM_INDEX = 0
START_INDEX = 1
STOP_INDEX = 2


def get_d4tools_intervals_completeness(
    d4_file_path: str, bed_file_path: str, completeness_thresholds: List[int]
) -> List[Dict]:
    """Return coverage completeness over all intervals of a bed file using the perc_cov d4tools command."""
    threshold_stats = []
    d4tools_stats_perc_cov: str = subprocess.check_output(
        [
            "d4tools",
            "stat",
            "-s",
            f"perc_cov={','.join(str(threshold) for threshold in completeness_thresholds)}",
            "--region",
            bed_file_path,
            d4_file_path,
        ],
        text=True,
    )
    for line in d4tools_stats_perc_cov.splitlines():
        stats_dict: Dict = dict(
            (
                zip(
                    completeness_thresholds,
                    [float(stat) for stat in line.rstrip().split("\t")[3:]],
                )
            )
        )
        threshold_stats.append(stats_dict)

    return threshold_stats


def get_completeness_stats(
    d4_file_path: str,
    thresholds: List[int],
    interval_ids_coords: List[Tuple[str, tuple]],
    chrom_prefix: str,
) -> Dict[str, dict]:
    """Compute coverage and completeness over the given intervals of a d4 file."""

    interval_id_completeness_stats: Dict[str:dict] = {}

    bed_lines = [
        f"{chrom_prefix}{coords[CHROM_INDEX]}\t{coords[START_INDEX]}\t{coords[STOP_INDEX]}"
        for _, coords in interval_ids_coords
    ]

    # Write genomic intervals to a temporary file
    with tempfile.NamedTemporaryFile(mode="w") as intervals_bed:

        intervals_bed.write("\n".join(bed_lines))
        intervals_bed.flush()
        intervals_completeness = get_d4tools_intervals_completeness(
            d4_file_path=d4_file_path,
            bed_file_path=intervals_bed.name,
            completeness_thresholds=thresholds,
        )

    for index, interval_id_coord in enumerate(interval_ids_coords):
        interval_id_completeness_stats[interval_id_coord[0]] = intervals_completeness[
            index
        ]

    return interval_id_completeness_stats



from typing import Iterator, List

from fastapi import APIRouter, BackgroundTasks, Depends, HTTPException, Response, status
from fastapi.responses import JSONResponse
from sqlalchemy.orm import Session

from chanjo2.constants import MULTIPLE_PARAMS_NOT_SUPPORTED_MSG
from chanjo2.crud.intervals import get_gene_intervals, get_genes, get_interval_counts
from chanjo2.dbutil import get_session
from chanjo2.meta.handle_load_intervals import update_interval_table
from chanjo2.models import SQLExon, SQLTranscript
from chanjo2.models.pydantic_models import (
    Builds,
    ExonBase,
    GeneBase,
    GeneIntervalQuery,
    GeneQuery,
    IntervalType,
    TranscriptBase,
)

router = APIRouter()


def count_nr_filters(filters: List[str]) -> int:
    """Count the items in a query list that aren't null."""
    return sum(filter is not None for filter in filters)


@router.post("/intervals/load/genes/{build}")
def load_genes(
    background_tasks: BackgroundTasks,
    build: Builds,
    file_path: str,
    session: Session = Depends(get_session),
) -> Response:
    """Load genes in the given genome build."""

    print(f"Loading {build} genes.")
    background_tasks.add_task(
        update_interval_table, IntervalType.GENES, build, file_path, session
    )
    return JSONResponse(
        content={
            "detail": "Genes will be updated in background. Please check their availability in a few minutes."
        }
    )


@router.post("/intervals/genes", response_model=List[GeneBase])
async def genes(query: GeneQuery, session: Session = Depends(get_session)):
    """Return genes according to query parameters."""
    nr_filters = count_nr_filters(
        filters=[query.ensembl_ids, query.hgnc_ids, query.hgnc_symbols]
    )
    if nr_filters > 1:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=MULTIPLE_PARAMS_NOT_SUPPORTED_MSG,
        )

    return get_genes(
        db=session,
        build=query.build,
        ensembl_ids=query.ensembl_ids,
        hgnc_ids=query.hgnc_ids,
        hgnc_symbols=query.hgnc_symbols,
        limit=query.limit if nr_filters == 0 else None,
    )


@router.post("/intervals/load/transcripts/{build}", response_model=List[GeneBase])
def load_transcripts(
    background_tasks: BackgroundTasks,
    build: Builds,
    file_path: str,
    session: Session = Depends(get_session),
) -> Response:
    """Load transcripts in the given genome build."""

    print(f"Loading {build} transcripts.")
    background_tasks.add_task(
        update_interval_table, IntervalType.TRANSCRIPTS, build, file_path, session
    )
    return JSONResponse(
        content={
            "detail": "Transcripts will be updated in background. Please check their availability in a few minutes."
        }
    )


@router.post("/intervals/transcripts", response_model=List[TranscriptBase])
async def transcripts(
    query: GeneIntervalQuery, session: Session = Depends(get_session)
):
    """Return transcripts according to query parameters."""
    nr_filters = count_nr_filters(
        filters=[
            query.ensembl_ids,
            query.hgnc_ids,
            query.hgnc_symbols,
            query.ensembl_gene_ids,
        ]
    )
    if nr_filters > 1:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=MULTIPLE_PARAMS_NOT_SUPPORTED_MSG,
        )
    return get_gene_intervals(
        db=session,
        build=query.build,
        ensembl_ids=query.ensembl_ids,
        hgnc_ids=query.hgnc_ids,
        hgnc_symbols=query.hgnc_symbols,
        ensembl_gene_ids=query.ensembl_gene_ids,
        limit=query.limit if nr_filters == 0 else None,
        interval_type=SQLTranscript,
    )


@router.post("/intervals/load/exons/{build}")
def load_exons(
    background_tasks: BackgroundTasks,
    build: Builds,
    file_path: str,
    session: Session = Depends(get_session),
) -> Response:
    """Load exons in the given genome build."""

    print(f"Loading {build} exons.")
    background_tasks.add_task(
        update_interval_table, IntervalType.EXONS, build, file_path, session
    )
    return JSONResponse(
        content={
            "detail": "Exons will be updated in background. Please check their availability in a few minutes."
        }
    )


@router.post("/intervals/exons", response_model=List[ExonBase])
async def exons(query: GeneIntervalQuery, session: Session = Depends(get_session)):
    """Return exons in the given genome build."""
    nr_filters = count_nr_filters(
        filters=[
            query.ensembl_ids,
            query.hgnc_ids,
            query.hgnc_symbols,
            query.ensembl_gene_ids,
        ]
    )
    if nr_filters > 1:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=MULTIPLE_PARAMS_NOT_SUPPORTED_MSG,
        )

    return get_gene_intervals(
        db=session,
        build=query.build,
        ensembl_ids=query.ensembl_ids,
        hgnc_ids=query.hgnc_ids,
        hgnc_symbols=query.hgnc_symbols,
        ensembl_gene_ids=query.ensembl_gene_ids,
        limit=query.limit if nr_filters == 0 else None,
        interval_type=SQLExon,
    )


@router.get("/intervals/intervals_count_by_build", response_model=dict)
def intervals_count_by_build(session: Session = Depends(get_session)):
    """Returns the number of genes, transcripts and exons available in the database for each genome build."""
    return get_interval_counts(db=session)



import logging

import coloredlogs
from fastapi.applications import FastAPI


def configure_log(log: logging.Logger, app: FastAPI):
    """Configure logging."""

    current_log_level = log.getEffectiveLevel()
    logging.getLogger("sqlalchemy.engine.Engine").disabled = True
    coloredlogs.install(level="DEBUG" if app.debug else current_log_level)



import logging
import time
from os.path import isfile
from statistics import mean
from typing import Dict, List, Tuple

from fastapi import APIRouter, Depends, HTTPException, status
from sqlalchemy.orm import Session

from chanjo2.auth import get_current_user
from chanjo2.constants import WRONG_BED_FILE_MSG, WRONG_COVERAGE_FILE_MSG
from chanjo2.crud.intervals import get_genes, set_sql_intervals
from chanjo2.dbutil import get_session
from chanjo2.meta.handle_bed import (
    bed_file_interval_id_coords,
    sort_interval_ids_coords,
)
from chanjo2.meta.handle_completeness_stats import get_completeness_stats
from chanjo2.meta.handle_coverage_stats import (
    get_chromosomes_prefix,
    get_d4tools_chromosome_mean_coverage,
    get_d4tools_intervals_mean_coverage,
)
from chanjo2.meta.handle_d4 import get_samples_sex_metrics, set_interval_ids_coords
from chanjo2.meta.handle_report_contents import INTERVAL_TYPE_SQL_TYPE, get_mean
from chanjo2.models import SQLGene
from chanjo2.models.pydantic_models import (
    CoverageSummaryQuery,
    FileCoverageIntervalsFileQuery,
    FileCoverageQuery,
    IntervalCoverage,
    IntervalType,
    TranscriptTag,
    is_valid_url,
)

router = APIRouter()
LOG = logging.getLogger(__name__)


@router.post("/coverage/d4/interval/", response_model=IntervalCoverage)
def d4_interval_coverage(
    query: FileCoverageQuery, user: dict = Depends(get_current_user)
):
    """Return coverage on the given interval for a D4 resource located on the disk or on a remote server."""

    chrom_prefix: str = get_chromosomes_prefix(query.coverage_file_path)
    chrom: str = query.chromosome.replace("chr", "")
    chromosome = f"{chrom_prefix}{chrom}"

    if None in [query.start, query.end]:  # Coverage over an entire chromosome
        return IntervalCoverage(
            mean_coverage=get_d4tools_chromosome_mean_coverage(
                d4_file_path=query.coverage_file_path,
                chromosomes=[chromosome],
            )[0][1],
            completeness={},
            interval_id=chromosome,
        )

    interval_ids_coords = [
        (
            f"{chromosome}:{query.start}-{query.end}",
            (chromosome, query.start, query.end),
        )
    ]

    mean_coverage: float = get_d4tools_intervals_mean_coverage(
        d4_file_path=query.coverage_file_path,
        interval_ids_coords=interval_ids_coords,
        chrom_prefix=chrom_prefix,
    )[0]

    completeness_stats = get_completeness_stats(
        d4_file_path=query.coverage_file_path,
        interval_ids_coords=[(chrom, (chrom, query.start, query.end))],
        thresholds=query.completeness_thresholds,
        chrom_prefix=chrom_prefix,
    )

    return IntervalCoverage(
        mean_coverage=mean_coverage,
        completeness=completeness_stats[chrom],
        interval_id=f"{chromosome}:{query.start}-{query.end}",
    )


@router.post("/coverage/d4/interval_file/", response_model=List[IntervalCoverage])
def d4_intervals_coverage(
    query: FileCoverageIntervalsFileQuery, user: dict = Depends(get_current_user)
):
    """Return coverage on the given intervals for a D4 resource located on the disk or on a remote server."""

    start_time = time.time()

    if isfile(query.intervals_bed_path) is False:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=WRONG_BED_FILE_MSG,
        )

    interval_ids_coords: List[Tuple[str, Tuple[str, int, int]]] = (
        bed_file_interval_id_coords(file_path=query.intervals_bed_path)
    )

    interval_ids_coords = sort_interval_ids_coords(interval_ids_coords)

    chrom_prefix: str = get_chromosomes_prefix(query.coverage_file_path)
    intervals_coverage = get_d4tools_intervals_mean_coverage(
        d4_file_path=query.coverage_file_path,
        interval_ids_coords=interval_ids_coords,
        chrom_prefix=chrom_prefix,
    )

    intervals_completeness: Dict[str, Dict[int, float]] = get_completeness_stats(
        d4_file_path=query.coverage_file_path,
        thresholds=query.completeness_thresholds,
        interval_ids_coords=interval_ids_coords,
        chrom_prefix=chrom_prefix,
    )

    results: List[IntervalCoverage] = []
    for counter, interval_data in enumerate(interval_ids_coords):
        coords = f"{interval_data[1][0]}:{interval_data[1][1]}-{interval_data[1][2]}"
        interval_coverage = {
            "interval_type": IntervalType.CUSTOM,
            "interval_id": coords,
            "mean_coverage": intervals_coverage[counter],
            "completeness": intervals_completeness[interval_data[0]],
        }
        results.append(IntervalCoverage.model_validate(interval_coverage))

    LOG.debug(
        f"Time to compute stats on {counter+1} intervals and {len(query.completeness_thresholds)} coverage thresholds: {time.time() - start_time} seconds."
    )

    return results


@router.post("/coverage/d4/genes/summary", response_model=Dict)
def d4_genes_condensed_summary(
    query: CoverageSummaryQuery,
    db: Session = Depends(get_session),
    user: dict = Depends(get_current_user),
):
    """Returning condensed summary containing only sample's mean coverage and completeness above a default threshold."""

    condensed_stats = {}
    genes: List[SQLGene] = get_genes(
        db=db,
        build=query.build,
        ensembl_ids=None,
        hgnc_ids=query.hgnc_gene_ids,
        hgnc_symbols=None,
        limit=None,
    )

    sql_intervals = set_sql_intervals(
        db=db,
        interval_type=INTERVAL_TYPE_SQL_TYPE[query.interval_type],
        genes=genes,
        transcript_tags=[TranscriptTag.REFSEQ_MRNA],
    )

    for sample in query.samples:
        # Make sure path to d4 files provided in the query exists
        if isfile(sample.coverage_file_path) is False:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=WRONG_COVERAGE_FILE_MSG,
            )

        interval_ids_coords: List[Tuple[str, Tuple[str, int, int]]] = (
            set_interval_ids_coords(sql_intervals=sql_intervals)
        )

        # Sort intervals by chrom, start & stop
        interval_ids_coords = sort_interval_ids_coords(interval_ids_coords)

        chrom_prefix: str = get_chromosomes_prefix(sample.coverage_file_path)

        # Compute mean coverage over genomic intervals
        genes_mean_coverage: List[float] = get_d4tools_intervals_mean_coverage(
            d4_file_path=sample.coverage_file_path,
            interval_ids_coords=interval_ids_coords,
            chrom_prefix=chrom_prefix,
        )
        # Compute coverage completeness over genomic intervals
        genes_coverage_completeness: Dict[str, dict] = get_completeness_stats(
            d4_file_path=sample.coverage_file_path,
            thresholds=[query.coverage_threshold],
            interval_ids_coords=interval_ids_coords,
            chrom_prefix=chrom_prefix,
        )
        genes_coverage_completeness_values: List[float] = [
            value[query.coverage_threshold] * 100
            for value in genes_coverage_completeness.values()
        ]

        condensed_stats[sample.name] = {
            "mean_coverage": get_mean(float_list=genes_mean_coverage),
            "coverage_completeness_percent": (
                round(mean(genes_coverage_completeness_values), 2)
                if genes_coverage_completeness_values
                else "NA"
            ),
        }

    return condensed_stats


@router.get("/coverage/samples/predicted_sex", response_model=Dict)
async def get_samples_predicted_sex(
    coverage_file_path: str, user: dict = Depends(get_current_user)
):
    """Return predicted sex for a sample given the coverage over its sex chromosomes."""
    if (
        isfile(coverage_file_path) is False
        and is_valid_url(query.coverage_file_path) is False
    ):
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=WRONG_COVERAGE_FILE_MSG,
        )
    return get_samples_sex_metrics(d4_file_path=coverage_file_path)



from os import path
from typing import Dict, Optional

from fastapi import APIRouter, Depends, Form, HTTPException, Request, status
from fastapi.encoders import jsonable_encoder
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates
from pydantic_core._pydantic_core import ValidationError
from sqlalchemy.orm import Session
from starlette.datastructures import FormData
from typing_extensions import Annotated

from chanjo2 import __version__
from chanjo2.auth import get_current_user
from chanjo2.constants import DEFAULT_COVERAGE_LEVEL
from chanjo2.dbutil import get_session
from chanjo2.demo import DEMO_COVERAGE_QUERY_FORM, DEMO_GENE_OVERVIEW_QUERY_FORM
from chanjo2.meta.handle_report_contents import (
    get_gene_overview_coverage_stats,
    get_mane_overview_coverage_stats,
    get_report_data,
)
from chanjo2.models.pydantic_models import (
    Builds,
    GeneCoverage,
    GeneReportForm,
    IntervalType,
    ReportQuery,
)


def get_templates_path() -> str:
    """Returns the absolute path to the templates folder of this app."""
    APP_ROOT: str = path.abspath(path.join(path.dirname(__file__), ".."))
    return path.join(APP_ROOT, "templates")


templates = Jinja2Templates(directory=get_templates_path())
router = APIRouter()


@router.get("/overview/demo", response_class=HTMLResponse)
async def demo_overview(request: Request, db: Session = Depends(get_session)):
    """Return a demo genes overview page over a list of genes for a list of samples."""

    overview_query = ReportQuery.as_form(DEMO_COVERAGE_QUERY_FORM)
    overview_content: dict = get_report_data(
        query=overview_query, session=db, is_overview=True
    )
    return templates.TemplateResponse(
        request=request,
        name="overview.html",
        context={
            "software_version": __version__,
            "extras": overview_content["extras"],
            "levels": overview_content["levels"],
            "incomplete_coverage_rows": overview_content["incomplete_coverage_rows"],
        },
    )


@router.post("/overview", response_class=HTMLResponse)
async def overview(
    request: Request,
    build=Annotated[Builds, Form(...)],
    samples=Annotated[str, Form(...)],
    interval_type=Annotated[IntervalType, Form(...)],
    completeness_thresholds=Annotated[Optional[str], Form(None)],
    ensembl_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_symbols=Annotated[Optional[str], Form(None)],
    default_level=Annotated[Optional[int], Form(DEFAULT_COVERAGE_LEVEL)],
    db: Session = Depends(get_session),
    user: dict = Depends(get_current_user),
):
    """Return the genes overview page over a list of genes for a list of samples."""
    try:
        overview_query = ReportQuery.as_form(await request.form())

    except ValidationError as ve:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=ve.json(),
        )

    overview_content: dict = get_report_data(
        query=overview_query, session=db, is_overview=True
    )
    return templates.TemplateResponse(
        request=request,
        name="overview.html",
        context={
            "software_version": __version__,
            "extras": overview_content["extras"],
            "levels": overview_content["levels"],
            "incomplete_coverage_rows": overview_content["incomplete_coverage_rows"],
        },
    )


@router.post("/gene_overview", response_class=HTMLResponse)
async def gene_overview(
    request: Request,
    db: Session = Depends(get_session),
    user: dict = Depends(get_current_user),
):
    """Returns coverage overview stats for a group of samples over genomic intervals of a single gene."""
    form_data: FormData = await request.form()
    form_dict: dict = jsonable_encoder(form_data)

    try:
        validated_form = GeneReportForm(**form_dict)
    except ValidationError as ve:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=ve.json(),
        )

    gene_overview_content: Dict[str, List[GeneCoverage]] = (
        get_gene_overview_coverage_stats(form_data=validated_form, session=db)
    )

    return templates.TemplateResponse(
        request=request, name="gene-overview.html", context=gene_overview_content
    )


@router.get("/gene_overview/demo", response_class=HTMLResponse)
async def demo_gene_overview(
    request: Request,
    db: Session = Depends(get_session),
):
    """Returns coverage overview stats for a group of samples over genomic intervals of a single demo gene."""
    validated_form = GeneReportForm(**DEMO_GENE_OVERVIEW_QUERY_FORM)

    gene_overview_content: Dict[str, List[GeneCoverage]] = (
        get_gene_overview_coverage_stats(form_data=validated_form, session=db)
    )

    return templates.TemplateResponse(
        request=request, name="gene-overview.html", context=gene_overview_content
    )


@router.get("/mane_overview/demo", response_class=HTMLResponse)
async def demo_mane_overview(
    request: Request,
    db: Session = Depends(get_session),
):
    """Returns coverage overview stats for a group of samples over MANE transcripts of a demo list of genes."""
    overview_query = ReportQuery.as_form(DEMO_COVERAGE_QUERY_FORM)
    overview_query.interval_type = IntervalType.TRANSCRIPTS
    overview_query.build = Builds.build_38

    return templates.TemplateResponse(
        request=request,
        name="mane-overview.html",
        context=get_mane_overview_coverage_stats(query=overview_query, session=db),
    )


@router.post("/mane_overview", response_class=HTMLResponse)
async def mane_overview(
    request: Request,
    build=Annotated[Builds, Form(Builds.build_38)],
    samples=Annotated[str, Form(...)],
    interval_type=Annotated[IntervalType, Form(IntervalType.TRANSCRIPTS)],
    completeness_thresholds=Annotated[Optional[str], Form(None)],
    ensembl_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_ids=Annotated[Optional[str], Form(None)],
    hgnc_gene_symbols=Annotated[Optional[str], Form(None)],
    default_level=Annotated[Optional[int], Form(DEFAULT_COVERAGE_LEVEL)],
    db: Session = Depends(get_session),
    user: dict = Depends(get_current_user),
):
    """Returns coverage overview stats for a group of samples over MANE transcripts of a list of genes."""
    try:
        overview_query = ReportQuery.as_form(await request.form())

    except ValidationError as ve:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=ve.json(),
        )

    return templates.TemplateResponse(
        request=request,
        name="mane-overview.html",
        context=get_mane_overview_coverage_stats(query=overview_query, session=db),
    )



import base64
import logging
import os
from typing import Any, Dict

import httpx
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPublicNumbers
from fastapi import HTTPException, Request, status
from jose import JWTError, jwt

LOG = logging.getLogger(__name__)

ALGORITHMS = ["RS256"]


def construct_rsa_key(key_data: Dict[str, str]) -> RSAPublicKey:
    """
    Constructs a public RSA key from JWKS `n` and `e`.
    """
    n = int.from_bytes(base64.urlsafe_b64decode(key_data["n"] + "=="), byteorder="big")
    e = int.from_bytes(base64.urlsafe_b64decode(key_data["e"] + "=="), byteorder="big")
    public_numbers = RSAPublicNumbers(e, n)
    return public_numbers.public_key(default_backend())


async def get_current_user(request: Request) -> Dict[str, Any]:
    """
    Verifies JWT token from cookies or request headers..
    """
    AUDIENCE = os.environ.get("AUDIENCE")
    JWKS_URL = os.environ.get("JWKS_URL")

    if not JWKS_URL or not AUDIENCE:
        # Skip verification in dev mode
        return {"sub": "anonymous", "role": "dev", "auth_skipped": True}

    # Try to get token from Authorization header first
    auth_header = request.headers.get("Authorization")

    if auth_header and auth_header.startswith("Bearer "):
        token = auth_header.removeprefix("Bearer ").strip()
    else:
        # Fallback to token in cookies
        token = request.cookies.get("access_token")

    if not token:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED, detail="Missing token"
        )

    try:
        # Fetch JWKS
        resp = await httpx.AsyncClient().get(JWKS_URL)
        jwks = resp.json()

        # Get the key that matches the token's `kid`
        unverified_header = jwt.get_unverified_header(token)
        key = next(
            (k for k in jwks["keys"] if k["kid"] == unverified_header["kid"]), None
        )

        if not key:
            raise HTTPException(
                status_code=401, detail="Unable to find matching key in JWKS"
            )

        public_key: RSAPublicKey = construct_rsa_key(key)

        # Decode and validate token
        payload = jwt.decode(
            token,
            public_key,
            algorithms=ALGORITHMS,
            audience=AUDIENCE,
        )
        return payload

    except JWTError as e:
        raise HTTPException(status_code=401, detail=f"Token validation failed: {e}")
    except Exception as e:
        raise HTTPException(status_code=500, detail=f"Unexpected error: {e}")



from dataclasses import dataclass

from sqlalchemy import JSON, Column, Enum, ForeignKey, Index, Integer, String

from chanjo2.dbutil import Base
from chanjo2.models.pydantic_models import Builds


@dataclass
class Interval(Base):
    """Used to define a single genomic interval."""

    __tablename__ = "intervals"

    id = Column(Integer, primary_key=True, index=True)
    chromosome = Column(String(6), nullable=False)
    start = Column(Integer, nullable=False)
    stop = Column(Integer, nullable=False)


@dataclass
class Gene(Base):
    """Used to define a gene entity."""

    __tablename__ = "genes"
    id = Column(Integer, primary_key=True, index=True)
    chromosome = Column(String(6), nullable=False)
    start = Column(Integer, nullable=False)
    stop = Column(Integer, nullable=False)
    ensembl_ids = Column(JSON, nullable=False)
    hgnc_id = Column(Integer, nullable=True, index=True)
    hgnc_symbol = Column(String(64), nullable=True)
    build = Column(
        Enum(Builds, values_callable=lambda x: Builds.get_enum_values()), index=True
    )
    __table_args__ = (
        Index("gene_idx_hgnc_id_build", "hgnc_id", "build"),
        Index("gene_idx_hgnc_symbol_build", "hgnc_symbol", "build"),
    )


@dataclass
class Transcript(Base):
    """Used to define a transcript entity."""

    __tablename__ = "transcripts"
    id = Column(Integer, primary_key=True, index=True)
    chromosome = Column(String(6), nullable=False)
    start = Column(Integer, nullable=False)
    stop = Column(Integer, nullable=False)
    ensembl_id = Column(String(24), nullable=False, index=True)
    refseq_mrna = Column(String(24), nullable=True)
    refseq_mrna_pred = Column(String(24), nullable=True)
    refseq_ncrna = Column(String(24), nullable=True)
    refseq_mane_select = Column(String(24), nullable=True, index=True)
    refseq_mane_plus_clinical = Column(String(24), nullable=True, index=True)
    ensembl_gene_id = Column(String(24), nullable=False, index=True)
    build = Column(
        Enum(Builds, values_callable=lambda x: Builds.get_enum_values()), index=True
    )

    __table_args__ = (
        Index("ensembl_gene_build_id", "ensembl_gene_id", "build", "ensembl_id"),
    )


@dataclass
class Exon(Base):
    """Used to define an exon entity."""

    __tablename__ = "exons"
    id = Column(Integer, primary_key=True, index=True)
    chromosome = Column(String(6), nullable=False)
    start = Column(Integer, nullable=False)
    stop = Column(Integer, nullable=False)
    rank_in_transcript = Column(Integer, nullable=False)
    ensembl_id = Column(String(24), nullable=False)
    ensembl_transcript_id = Column(String(24), nullable=False, index=True)
    ensembl_gene_id = Column(String(24), nullable=False, index=True)
    build = Column(
        Enum(Builds, values_callable=lambda x: Builds.get_enum_values()), index=True
    )

    __table_args__ = (
        Index(
            "exon_idx_ensembl_gene_build_transcript",
            "ensembl_gene_id",
            "build",
            "ensembl_transcript_id",
        ),
    )
